import os
import re
import shutil
import subprocess

import necxxt.utils
import necxxt.utils.logging


def checkCoverage(config=None, coverage_info=None):
    if not config:
        config = necxxt.utils.readModule()

    if not coverage_info:
        coverage_info = os.path.join(os.getcwd(), "build", "coverage.info")

    if not os.path.isfile(coverage_info):
        necxxt.utils.logging.error(
            "Could not read coverage file \"{}\"".format(coverage_info))
        exit(2)

    lcov_exe = shutil.which("lcov")
    if not lcov_exe:
        necxxt.utils.logging.error("lcov is not installed")
        exit(3)

    cmd = [
        lcov_exe,
        "--rc",
        "lcov_branch_coverage=1",
        "--summary",
        coverage_info
    ]

    p = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
    output = p.communicate()[0].decode("utf-8")

    re_lines = re.compile(".*lines.*:\s*(.*)%.*")
    re_functions = re.compile(".*functions.*:\s*(.*)%.*")
    re_branches = re.compile(".*branches.*:\s*(.*)%.*")

    lines = None
    functions = None
    branches = None
    for line in output.split("\n"):
        match = re_lines.match(line)
        if match:
            lines = float(match.group(1))

        match = re_functions.match(line)
        if match:
            functions = float(match.group(1))

        match = re_branches.match(line)
        if match:
            branches = float(match.group(1))

    coverageThresholdsLines = config.get("coverage", {}).get(
        "thresholds", {}).get("lines", 0.0)
    coverageThresholdsFunctions = config.get(
        "coverage", {}).get("thresholds", {}).get("functions", 0.0)
    coverageThresholdsBranches = config.get(
        "coverage", {}).get("thresholds", {}).get("branches", 0.0)

    necxxt.utils.logging.debug("Current line coverage: {:.2f}%".format(lines))
    necxxt.utils.logging.debug(
        "Current function coverage: {:.2f}%".format(functions))
    necxxt.utils.logging.debug(
        "Current branch coverage: {:.2f}%".format(branches))

    return_code = 100

    if coverageThresholdsLines:
        if lines < coverageThresholdsLines:
            if return_code == 100:
                print()

            necxxt.utils.logging.error(
                "Line coverage dropped under {:.2f}% --> {:.2f}%".format(coverageThresholdsLines, lines))
            return_code += 1

    if coverageThresholdsFunctions:
        if functions < coverageThresholdsFunctions:
            if return_code == 100:
                print()

            necxxt.utils.logging.error(
                "Function coverage dropped under {:.2f}% --> {:.2f}%".format(coverageThresholdsFunctions, functions))
            return_code += 1

    if coverageThresholdsBranches:
        if branches < coverageThresholdsBranches:
            if return_code == 100:
                print()

            necxxt.utils.logging.error(
                "Branch coverage dropped under {:.2f}% --> {:.2f}%".format(coverageThresholdsBranches, branches))
            return_code += 1

    if return_code > 100:
        exit(return_code)



import json
import os

import necxxt.utils


def initialize():
    print("This utility will walk you through creating a necxxt.json file.")
    print("It only covers the most common items, and tries to guess sensible defaults.")
    print()
    print("Use `necxxt install <pkg>` afterwards to install a package and")
    print("save it as a dependency in the necxxt.json file.")
    print()
    print("Press ^C at any time to quit.")

    module = {
        "name": os.path.basename(os.path.normpath(os.getcwd())),
        "version": "1.0.0",
        "description": "",
        "author": "",
        "license": "MIT"
    }

    module["name"] = input("package name: ({}) ".format(
        module["name"])).strip() or module["name"]
    module["version"] = input("version: ({}) ".format(
        module["version"])).strip() or module["version"]
    module["description"] = input(
        "description: ").strip() or module["description"]
    git_repository = input("git repository: ").strip()
    if git_repository:
        module["repository"] = {
            "type": "git",
            "url": git_repository
        }
    module["author"] = input(
        "author: ").strip() or module["author"]
    module["license"] = input("license: ({}) ".format(
        module["license"])).strip() or module["license"]
    print("About to write to {}:".format(
        os.path.join(os.getcwd(), "necxxt.json")))
    print()
    print(json.dumps(module, indent=4))
    print()
    response = input("Is this OK? (yes) ") or "yes"
    if not response in ["yes"]:
        print("Aborted.")
        print()
        exit(1)

    necxxt.utils.writeModule(module)



import copy
import json
import os
import pathlib
import shutil
import subprocess

from .build import checkIfModulesAreInstalled, precompileModules, compileModules, compileSources, buildStatic
from .coverage import checkCoverage
from .init import initialize

import necxxt.dependencies


class Command:
    def help(self):
        raise NotImplementedError

    def run(self, *args, **kwargs):
        raise NotImplementedError

    def __str__(self):
        return self._name


class BuildCommand(Command):
    def help(self):
        print("Build a necxxt compatible project")

    def run(self, *args, **kwargs):
        if os.path.isdir("build"):
            shutil.rmtree("build")

        lock_path = os.path.join(os.getcwd(), "necxxt-lock.json")
        if not os.path.isfile(lock_path):
            necxxt.utils.logging.error(
                "ENOENT: no such file or directory, open '{}'".format(lock_path))
            exit(1)

        with open(lock_path, "r") as f:
            lock = json.load(f)

        modules = {}
        for dependency in lock["dependencies"].keys():
            requires = lock["dependencies"][dependency].get("requires", {})
            modules[dependency] = list(requires.keys())

        build_modules = copy.deepcopy(modules)

        order = []
        i = 0
        while i < len(modules.keys()):
            keys = list(modules.keys())
            key = keys[i]
            dependencies = modules[key]

            if len(dependencies) == 0:
                order.append(key)
                del modules[key]
                for d in modules.keys():
                    if key in modules[d]:
                        modules[d].remove(key)

            i += 1

            length = len(modules.keys())
            if length > 0 and i >= length:
                i = 0

        checkIfModulesAreInstalled(order)
        precompileModules(order, build_modules)
        compileModules(order, build_modules)
        compileSources(order, build_modules)
        buildStatic(order, build_modules)


class CheckCoverageCommand(Command):
    def help(self):
        print("Check the coverage of a project")

    def run(self, *args, **kwargs):
        config = None
        coverage_info = None
        for key, value in kwargs.items():
            if key == "config":
                config = necxxt.utils.readJson(value)
            elif key == "coverage-info":
                coverage_info = value

        checkCoverage(config=config, coverage_info=coverage_info)


class InitCommand(Command):
    def help(self):
        print("Initialize a necxxt project")

    def run(self, *args, **kwargs):
        initialize()


class InstallCommand(Command):
    def help(self):
        print("Install the dependencies of a necxxt compatible project")

    def run(self, *args, **kwargs):
        necxxt.dependencies.installDependencies()


class RunCommand(Command):
    def help(self):
        print("Run the executable of a necxxt compatible project")

    def run(self, *args, **kwargs):
        lock_path = os.path.join(os.getcwd(), "necxxt-lock.json")
        if not os.path.isfile(lock_path):
            necxxt.utils.logging.error(
                "ENOENT: no such file or directory, open '{}'".format(lock_path))
            exit(1)

        with open(lock_path, "r") as f:
            lock = json.load(f)

        exe = os.path.join(os.getcwd(), "build", lock.get("name", ""))
        if not os.path.isfile(exe):
            necxxt.utils.logging.error(
                "ENOENT: no such file or directory, open '{}'".format(exe))
            exit(1)

        subprocess.call([exe])
